from __future__ import annotations

import matplotlib as mpl
import matplotlib.pyplot as plt

from contextlib import contextmanager
from pathlib import Path


def get_rcParams(dpi: int = 600, hide_ticks: bool = False) -> dict:  # noqa: N802
    # Lines should be 1px
    # Linewidth is given in points
    # 1 Point = DPI/72 px
    # 1 px = 72 / DPI Points

    px_in_pts = 72 / dpi
    rcParams = {}  # noqa: N806
    rcParams["lines.linewidth"] = px_in_pts
    rcParams["lines.markersize"] = px_in_pts
    rcParams["patch.linewidth"] = px_in_pts
    rcParams["axes.linewidth"] = 2 * px_in_pts
    rcParams["xtick.major.width"] = px_in_pts
    rcParams["ytick.major.width"] = px_in_pts
    rcParams["font.size"] = 30 * px_in_pts

    rcParams["legend.frameon"] = False
    rcParams["legend.loc"] = "lower right"

    if hide_ticks:
        rcParams["axes.spines.left"] = False
        rcParams["axes.spines.right"] = False
        rcParams["axes.spines.bottom"] = False
        rcParams["axes.spines.top"] = False
        rcParams["xtick.bottom"] = False
        rcParams["ytick.left"] = False
        rcParams["xtick.labelbottom"] = False
        rcParams["ytick.labelleft"] = False

    return rcParams


@contextmanager
def create_plot_ax(
    plot_dir: str | Path, plot_name: str, dpi: int = 600, figw: int = 9, figh: int = 9
) -> plt.Axes:
    if not isinstance(plot_dir, (Path, str)):
        msg = "'plot_dir' must be str or Path"
        raise TypeError(msg)
    plot_dir = Path(plot_dir)
    if not plot_dir.exists():
        msg = "Given 'plot_dir' does not exist. Please create it first."
        raise FileNotFoundError(msg)

    with plt.style.context("default"):
        with mpl.rc_context():
            f, ax = plt.subplots(figsize=(figw, figh), dpi=dpi)

            try:
                yield ax
            finally:
                f.savefig(plot_dir / f"{plot_name}.png", bbox_inches="tight")
                plt.close(f)


@contextmanager
def create_scenario_plot_ax(  # noqa: PLR0912
    plot_dir: str | Path,
    plot_name: str,
    dpi: int = 600,
    figw: int = 9,
    figh: int = 9,
    hide_ticks: bool = False,
    hide_legend: bool = False,
    hide_title: bool = False,
    hide_xlabel: bool = False,
    hide_ylabel: bool = False,
    hide_grid: bool = False,
    clean: bool = False,
) -> plt.Axes:
    if not isinstance(plot_dir, (Path, str)):
        msg = "'plot_dir' must be str or Path"
        raise TypeError(msg)
    plot_dir = Path(plot_dir)
    if not plot_dir.exists():
        msg = "Given 'plot_dir' does not exist. Please create it first."
        raise FileNotFoundError(msg)

    if clean:
        hide_ticks = True
        hide_legend = True
        hide_title = True
        hide_xlabel = True
        hide_ylabel = True
        hide_grid = True

    if plot_dir:
        if not isinstance(plot_dir, (Path, str)):
            msg = "'plot_dir' must be str or Path"
            raise TypeError(msg)
        plot_dir = Path(plot_dir)
        if not plot_dir.exists():
            msg = "Given 'plot_dir' does not exist. Please create it first."
            raise FileNotFoundError(msg)

    with plt.style.context("default"):
        with mpl.rc_context(get_rcParams(dpi=dpi, hide_ticks=hide_ticks)):
            f, ax = plt.subplots(figsize=(figw, figh), dpi=dpi)

            try:
                yield ax

            finally:
                # Handle legend
                legend = ax.get_legend()
                if legend:
                    if hide_legend:
                        legend.remove()
                    else:
                        legend.set_bbox_to_anchor((1, 1))
                        legend.set_ncols(3)

                if hide_title:
                    ax.set_title("")

                if hide_xlabel:
                    ax.set_xlabel("")

                if hide_ylabel:
                    ax.set_ylabel("")

                if hide_grid:
                    ax.grid(False)

                ax.set_aspect("equal")

                # Save to file
                f.savefig(plot_dir / f"{plot_name}.png", bbox_inches="tight")
                plt.close(f)



import numpy as np

from .road_segment import RoadSegment


class ArcSegment(RoadSegment):
    def __init__(self, length: float = 100, radius: float = 900) -> None:
        """
        length: Length of the road segment in m
        radius: Curve radius of reference line in m.
        """
        self._config = {"length": length, "radius": radius}

        super().__init__()

        if radius < 900:
            msg = f"The radius of an arc segment must at least be 900m. Now it is: {radius}"
            raise ValueError(msg)

        self._length = length
        self._radius = radius

        self._center_x = None
        self._center_y = None
        self._start_angle = None
        self._end_angle = None
        self._sampling_angle = None

        self._curvature = None

    def __str__(self) -> str:
        return (
            f"<{self.__class__.__name__} length={self._length}, radius={self._radius}>"
        )

    @property
    def config(self) -> dict:
        return self._config

    @property
    def length(self) -> float:
        return self._length

    @property
    def radius(self) -> float:
        return self._radius

    @property
    def curvature(self) -> float:
        if self._curvature is None:
            msg = (
                "Need to call compute_ref_line() first by adding the segment to a road."
            )
            raise RuntimeError(msg)

        return self._curvature

    @staticmethod
    def create_arc(
        center_x: float,
        center_y: float,
        radius: float,
        start_angle: float,
        end_angle: float,
        sampling_angle: float = np.rad2deg(1),
    ) -> np.ndarray:
        angles = np.arange(start_angle, end_angle + sampling_angle, sampling_angle)

        x = center_x + radius * np.cos(angles)
        y = center_y + radius * np.sin(angles)

        arc_points = np.array([x, y]).transpose()

        return arc_points

    def compute_ref_line(
        self, start_x: float, start_y: float, start_heading: float
    ) -> np.ndarray:
        full_circle_distance = 2 * np.pi * self._radius
        angle_dist = (self._length / full_circle_distance) * 2 * np.pi
        sampling_angle = (self._sampling_distance / full_circle_distance) * 2 * np.pi

        center_x = start_x + self._radius * np.cos(start_heading + np.pi / 2)
        center_y = start_y + self._radius * np.sin(start_heading + np.pi / 2)

        start_angle = (-np.pi / 2) + start_heading
        end_angle = start_angle + angle_dist

        ref_line = ArcSegment.create_arc(
            center_x, center_y, self._radius, start_angle, end_angle, sampling_angle
        )

        self._center_x = center_x
        self._center_y = center_y
        self._start_angle = start_angle
        self._end_angle = end_angle
        self._sampling_angle = sampling_angle

        self._curvature = 1 / self._radius

        self._ref_line = ref_line
        self._ref_line_curvature = self._curvature * np.ones_like(ref_line[:, 0])
        self._ref_line_heading = (
            np.arange(start_angle, end_angle + sampling_angle, sampling_angle)
            + np.pi / 2
        )

        return ref_line

    def compute_offset_line(self, lateral_offset: float) -> np.ndarray:
        offset_line = ArcSegment.create_arc(
            self._center_x,
            self._center_y,
            self._radius + lateral_offset,
            self._start_angle,
            self._end_angle,
            self._sampling_angle,
        )

        self._offset_lines[lateral_offset] = offset_line

        return offset_line



from __future__ import annotations

import lanelet2

import numpy as np
from matplotlib.patches import Rectangle
from typing import TYPE_CHECKING
from vehiclemodels.parameters_vehicle1 import parameters_vehicle1
from vehiclemodels.parameters_vehicle2 import parameters_vehicle2
from vehiclemodels.parameters_vehicle3 import parameters_vehicle3

from .rendering import Renderable

if TYPE_CHECKING:
    import matplotlib.pyplot as plt
    from omegaconf import DictConfig
    from .road import Road


class Vehicle(Renderable):
    VEHICLE_TYPE_PARAMETERS = {
        "small": parameters_vehicle1,  # VehicleType.FORD_ESCORT,
        "medium": parameters_vehicle2,  # VehicleType.BMW_320i,
        "van": parameters_vehicle3,  # VehicleType.VW_VANAGON
        # "semi_trailer": 4  # Currently not available
        "custom": None,  # Take width and length from parameters
    }

    ACCELERATION_PROFILES = ["constant"]

    LC_TYPES = ("polynomial", "vy")

    def __init__(
        self,
        vehicle_id: int,
        lanelet_id: int,
        s0: float,
        t0: float,
        v0: float,
        a0: float = 0,
        a_delay: float = 0,
        a_profile: str = "constant",
        lc_direction: int = 0,
        lc_delay: float = 0,
        lc_duration: float = 3,
        lc_type: str = "polynomial",
        lc_vy: float = 0,
        inverse_driving_direction: bool = False,
        vehicle_type_name: str = "medium",
        length: float | None = None,
        width: float | None = None,
        from_data: bool = False,
    ) -> None:
        """
        lanelet_id: ID of initial lanelet the vehicle is starting on
        s0: Initial longitudinal position relative to the given lanelet in m
        t0: Initial lateral offset to the center line of the given lanelet in m
        v0: Initial speed in m/s
        a: Acceleration in m/s^2
        a_delay: Time to wait before applying acceleration
        a_profile: How the acceleration changes over time
        lc_direction: Direction of lane change. 0: no lane change, 1: left, -1: right
        lc_delay: Time to wait before lane change
        lc_duration: Time the lane change takes to be completed
        lc_type: How the lc is done
        lc_vy: Only used if lc_type is "vy"
        vehicle_type_name: Type of the vehicle
        """

        if a_profile not in self.ACCELERATION_PROFILES:
            msg = f"Unsupported acceleration profile specifed: {a_profile}. Choose from: {self.ACCELERATION_PROFILES}."
            raise AttributeError(msg)

        if lc_direction not in (-1, 0, 1):
            msg = "lc_direction can be -1, 0, or 1"
            raise ValueError(msg)

        if lc_type not in self.LC_TYPES:
            msg = f"Unsupported lc type specifed: {lc_type}. Choose from: {self.LC_TYPES}."
            raise AttributeError(msg)

        if vehicle_type_name not in self.VEHICLE_TYPE_PARAMETERS:
            msg = "Vehicle type is not available"
            raise ValueError(msg)

        if vehicle_type_name == "custom" and length is None and width is None:
            msg = "If vehicle type is 'custom', parameters 'length' and 'width' must be not None"
            raise ValueError(msg)

        if vehicle_type_name != "custom" and length is not None and width is not None:
            msg = "If vehicle type is not 'custom', parameters 'length' and 'width' must be None"
            raise ValueError(msg)

        self._config = {
            "vehicle_id": vehicle_id,
            "lanelet_id": lanelet_id,
            "s0": s0,
            "t0": t0,
            "v0": v0,
            "a0": a0,
            "a_delay": a_delay,
            "a_profile": a_profile,
            "lc_direction": lc_direction,
            "lc_delay": lc_delay,
            "lc_duration": lc_duration,
            "lc_type": lc_type,
            "lc_vy": lc_vy,
            "vehicle_type_name": vehicle_type_name,
        }

        self._vehicle_id = vehicle_id
        self._lanelet_id = lanelet_id
        self._s0 = s0
        self._t0 = t0
        self._v0 = v0

        self._a0 = a0
        self._a_delay = a_delay
        self._a_profile = a_profile

        self._lc_direction = lc_direction
        self._lc_delay = lc_delay
        self._lc_duration = lc_duration
        self._lc_type = lc_type
        self._lc_vy = lc_vy

        self._inverse_driving_direction = inverse_driving_direction

        self._vehicle_type_name = vehicle_type_name
        self._vehicle_parameters = None
        if self._vehicle_type_name == "custom":
            self._length = length
            self._width = width
        else:
            self._vehicle_parameters = self.VEHICLE_TYPE_PARAMETERS[
                self._vehicle_type_name
            ]()
            self._length = self._vehicle_parameters.l
            self._width = self._vehicle_parameters.w

        # Compiled values
        self._compiled = False
        self._x = None
        self._y = None
        self._heading = None
        self._v = None
        self._a = None

        # If the vehicle is created directly from data, this is True
        self._initialized_from_data = from_data

    @classmethod
    def from_data(
        cls,
        vehicle_id: int,
        lanelet_id: int,
        x: np.ndarray,
        y: np.ndarray,
        heading: np.ndarray,
        v: np.ndarray,
        a: np.ndarray,
        vehicle_type_name: str = "medium",
        length: float | None = None,
        width: float | None = None,
    ) -> Vehicle:
        if x.shape[0] == 0:
            msg = "All data arrays must have some values."
            raise ValueError(msg)

        if len(x.shape) != 1:
            msg = "All data arrays must be of shape (X,)."
            raise ValueError(msg)

        if not (x.shape == y.shape == heading.shape == v.shape == a.shape):
            msg = "All data arrays must have same shape"
            raise ValueError(msg)

        vehicle = Vehicle(
            vehicle_id,
            lanelet_id,
            None,
            None,
            heading[0],
            v[0],
            vehicle_type_name=vehicle_type_name,
            length=length,
            width=width,
            from_data=True,
        )

        vehicle.set_data(x, y, heading, v, a)

        return vehicle

    @property
    def config(self) -> dict:
        return self._config

    @property
    def id(self) -> int:
        return self._vehicle_id

    @property
    def vehicle_parameters(self) -> DictConfig:
        if self._vehicle_type_name == "custom":
            msg = (
                "If vehicle_type_name is 'custom', vehicle_parameters are not available"
            )
            raise ValueError(msg)
        return self._vehicle_parameters

    @property
    def lanelet_id(self) -> int:
        return self._lanelet_id

    @property
    def s0(self) -> float:
        if self._initialized_from_data:
            msg = "Not possible to access s0 when initialized from data."
            raise Exception(msg)
        return self._s0

    @property
    def t0(self) -> float:
        if self._initialized_from_data:
            msg = "Not possible to access t0 when initialized from data."
            raise Exception(msg)
        return self._t0

    @property
    def v0(self) -> float:
        return self._v0

    @property
    def length(self) -> float:
        return self._length

    @property
    def width(self) -> float:
        return self._width

    @property
    def x(self) -> np.ndarray:
        if not self._compiled:
            msg = "Please call .compile() before accessing this data."
            raise Exception(msg)
        return self._x

    @property
    def y(self) -> np.ndarray:
        if not self._compiled:
            msg = "Please call .compile() before accessing this data."
            raise Exception(msg)
        return self._y

    @property
    def heading(self) -> np.ndarray:
        if not self._compiled:
            msg = "Please call .compile() before accessing this data."
            raise Exception(msg)
        return self._heading

    @property
    def v(self) -> np.ndarray:
        if not self._compiled:
            msg = "Please call .compile() before accessing this data."
            raise Exception(msg)
        return self._v

    @property
    def a(self) -> np.ndarray:
        if not self._compiled:
            msg = "Please call .compile() before accessing this data."
            raise Exception(msg)
        return self._a

    @property
    def inverse_driving_direction(self) -> bool:
        return self._inverse_driving_direction

    @property
    def vehicle_type_name(self) -> str:
        return self._vehicle_type_name

    @property
    def initialized_from_data(self) -> bool:
        return self._initialized_from_data

    def set_data(
        self, x: np.array, y: np.array, heading: np.array, v: np.array, a: np.array
    ) -> None:
        if not self._initialized_from_data:
            msg = "Cannot set data unless the vehicle is initialized with 'from_data=True'."
            raise Exception(msg)

        self._x = x
        self._y = y
        self._heading = heading
        self._v = v
        self._a = a

        self._compiled = True

    @property
    def compiled(self) -> bool:
        return self._compiled

    def compile(self, road: Road, duration: float, dt: float) -> None:  # noqa: PLR0912
        """
        Create absolute cartesian coordinates etc
        """

        # NEED LANELET2 MAP FIRST
        lanelet_map = road.lanelet_map

        # Find lanelet by id
        lanelet = lanelet_map.laneletLayer[self._lanelet_id]

        n_steps = self._n_steps_from_duration_dt(duration, dt)

        # a
        acceleration_vector = self._get_full_acceleration_vector(n_steps, dt)
        speed_change_vector = acceleration_vector * dt

        # v
        speed_vector = self._v0 * np.ones_like(speed_change_vector)
        speed_vector[1:] = speed_vector[1:] + np.cumsum(speed_change_vector)[:-1]
        speed_vector[speed_vector < 1e-2] = 0.0

        # s
        driving_direction = -1 if self._inverse_driving_direction else 1
        lon_position_change_vector = driving_direction * speed_vector * dt
        lon_position_vector = self._s0 * np.ones_like(speed_change_vector)
        lon_position_vector[1:] = (
            lon_position_vector[1:] + np.cumsum(lon_position_change_vector)[:-1]
        )

        # t
        lat_offset_vector = self._t0 * np.ones((n_steps,))
        lc_start_step = int(self._lc_delay / dt)

        # Lane change
        if self._lc_direction in (-1, 1):
            if self._lc_direction == -1:
                # Assumption: right lanelet ID is one less than initial lanelet id
                target_lanelet_id = self._lanelet_id - 1

                if target_lanelet_id not in lanelet_map.laneletLayer:
                    msg = "Lane change to the right not possible. No lanelet to the right."
                    raise ValueError(msg)

                target_lanelet = lanelet_map.laneletLayer[target_lanelet_id]

                if not lanelet2.geometry.rightOf(target_lanelet, lanelet):
                    msg = "Right lanelet has another ID than expected."
                    raise Exception(msg)

            elif self._lc_direction == 1:
                # Assumption: right lanelet ID is one more than initial lanelet id
                target_lanelet_id = self._lanelet_id + 1

                if target_lanelet_id not in lanelet_map.laneletLayer:
                    msg = (
                        "Lane change to the left not possible. No lanelet to the left."
                    )
                    raise ValueError(msg)

                target_lanelet = lanelet_map.laneletLayer[target_lanelet_id]

                if not lanelet2.geometry.leftOf(target_lanelet, lanelet):
                    msg = "Left lanelet has another ID than expected."
                    raise Exception(msg)

            if self._lc_type == "polynomial":
                lc_traj = self._generate_lc_trajectory(
                    road,
                    lon_position_vector[lc_start_step],
                    speed_vector[lc_start_step],
                    self._lanelet_id,
                    target_lanelet_id,
                    self._lc_duration,
                    dt,
                )

                n_lc_steps = lc_traj.shape[0]

                lat_offset_vector[lc_start_step : lc_start_step + n_lc_steps] = lc_traj[
                    :, 1
                ]
                lat_offset_vector[lc_start_step + n_lc_steps :] = lc_traj[-1, 1]

            elif self._lc_type == "vy":
                lat_offset_change_vector = (
                    self._lc_vy
                    * dt
                    * np.ones_like(lat_offset_vector)
                    * self._lc_direction
                )
                lat_offset_vector[1:] = (
                    lat_offset_vector[1:] + np.cumsum(lat_offset_change_vector)[:-1]
                )

                # End of LC
                # Find max possible t positon
                initial_llt = road.lanelet_map.laneletLayer[self._lanelet_id]
                target_llt = road.lanelet_map.laneletLayer[target_lanelet_id]
                x_lc1, y_lc1 = road.from_frenet_to_cart(
                    target_llt.centerline, self._s0, 0
                )
                _, t_lc1 = road.from_cart_to_frenet(
                    initial_llt.centerline, x_lc1, y_lc1
                )
                max_t = t_lc1  # - self.width / 2

                if self._lc_direction == 1:
                    lat_offset_vector[lat_offset_vector > max_t] = max_t
                else:
                    lat_offset_vector[lat_offset_vector < max_t] = max_t

        # Positions to cartesian
        x, y = road.from_frenet_to_cart(
            lanelet.centerline, lon_position_vector, lat_offset_vector
        )

        # Heading, Assumption: Heading does not change in last time step
        if self._lc_type == "vy":
            if driving_direction == 1:
                heading = np.zeros_like(acceleration_vector)
            elif driving_direction == -1:
                heading = np.pi * np.ones_like(acceleration_vector)
            else:
                raise RuntimeError
        else:
            # Get initial heading guess from the road
            x_diff0, y_diff0 = road.from_frenet_to_cart(
                lanelet.centerline,
                np.array([lon_position_vector[0], lon_position_vector[0] + 0.5]),
                np.zeros((2,)),
            )
            initial_heading_guess = np.arctan2(np.diff(y_diff0), np.diff(x_diff0))
            # Handle standstills (heading remains constant)
            standstill = np.isclose(
                np.diff(x, append=x[-1]) + np.diff(y, append=y[-1]), 0, atol=1e-3
            )
            if np.all(standstill):
                heading = initial_heading_guess * np.ones_like(x)
            else:
                heading = np.arctan2(np.diff(y, append=y[-1]), np.diff(x, append=x[-1]))

                standstill_start_idcs = np.where(
                    np.diff(standstill.astype(int), prepend=0) == 1
                )[0]
                standstill_end_idcs = np.where(
                    np.diff(standstill.astype(int), append=0) == -1
                )[0]
                for start_idx, end_idx in zip(
                    standstill_start_idcs, standstill_end_idcs
                ):
                    if start_idx == 0:
                        value = initial_heading_guess
                    else:
                        value = heading[start_idx - 1]
                    heading[start_idx : end_idx + 1] = value

        self._a = acceleration_vector
        self._v = speed_vector
        self._x = x
        self._y = y
        self._heading = heading

        self._compiled = True

    @staticmethod
    def _n_steps_from_duration_dt(duration: float, dt: float) -> int:
        n_steps = int(np.floor(duration / dt))
        return n_steps

    @staticmethod
    def _elapsed_time_from_n_steps_dt(n_steps: float, dt: float) -> float:
        elapsed_time = np.arange(n_steps, dt)
        return elapsed_time

    def _get_full_acceleration_vector(self, n_steps: int, dt: float) -> np.ndarray:
        if self._a_profile == "constant":
            acceleration_vector = self._a0 * np.ones((n_steps,))

            # Delay
            delay_steps = int(self._a_delay / dt)
            acceleration_vector[:delay_steps] = 0.0

        return acceleration_vector

    def _generate_lc_trajectory(
        self,
        road: Road,
        s_lc0: float,
        v0: float,
        llt_id_lc0: int,
        llt_id_lc1: int,
        lc_duration: float,
        dt: float,
    ) -> np.ndarray:
        """
        Return LC trajectory in frenet frame of llt_lc0 (before lc)
        """
        # !! COPIED FROM: https://gitlab.ika.rwth-aachen.de/lva/pilots/-/blob/main/pilots/highway_pilot/highway_pilot.py?ref_type=heads#L591
        # CHANGED!

        llt_lc0 = road.lanelet_map.laneletLayer[llt_id_lc0]
        llt_lc1 = road.lanelet_map.laneletLayer[llt_id_lc1]

        # Keep velocity, constant acceleration over lc_duration
        a = 0
        v1 = v0 + a * lc_duration

        # s position after lc
        s_lc1 = s_lc0 + ((v0 + v1) / 2) * lc_duration

        # Lateral positions in ref_frame
        x_lc0, y_lc0 = road.from_frenet_to_cart(llt_lc0.centerline, s_lc0, 0)
        x_lc1, y_lc1 = road.from_frenet_to_cart(llt_lc1.centerline, s_lc1, 0)

        _, t_lc0 = road.from_cart_to_frenet(llt_lc0.centerline, x_lc0, y_lc0)
        _, t_lc1 = road.from_cart_to_frenet(llt_lc0.centerline, x_lc1, y_lc1)

        # 5th order polynomial
        d = lc_duration
        p = np.array(
            [
                [0**5, 0**4, 0**3, 0**2, 0**1, 1],  # lateral start position
                [d**5, d**4, d**3, d**2, d**1, 1],  # lateral end positioin
                [0, 0, 0, 0, 1, 0],  # lateral start velocity
                [5 * d**4, 4 * d**3, 3 * d**2, 2 * d, 1, 0],  # lateral end velocity
                [0, 0, 0, 2, 0, 0],  # lateral start acceleration
                [20 * d**3, 12 * d**2, 6 * d, 2, 0, 0],
            ]
        )  # lateral end acceleration
        pv = np.linalg.solve(p, [t_lc0, t_lc1, 0.0, 0.0, 0, 0])

        trajectory_pts = []
        time_steps = np.arange(0, lc_duration + dt, dt)

        for t in time_steps:
            delta_s = s_lc0 + v0 * t + 0.5 * a * t * t
            delta_t = (
                pv[0] * t**5
                + pv[1] * t**4
                + pv[2] * t**3
                + pv[3] * t**2
                + pv[4] * t**1
                + pv[5]
            )
            trajectory_pts.append([delta_s, delta_t])

        trajectory_pts = np.array(trajectory_pts)

        return trajectory_pts

    def _plot_in_ax(
        self,
        ax: plt.Axes,
        timestep: int | None = None,
        include_trace: bool = False,
        color: str | None = None,
    ) -> None:
        if timestep and not (0 <= timestep < self._x.shape[0]):
            msg = "timestep is out of possible range."
            raise ValueError(msg)

        if color is None:
            color = "r"

        if timestep is None:
            x = self._x[0]
            y = self._y[0]
            heading = self._heading[0]
        else:
            x = self._x[timestep]
            y = self._y[timestep]
            heading = self._heading[timestep]

        rect_patch = Rectangle(
            (x - self.length / 2, y - self.width / 2),
            self.length,
            self.width,
            angle=np.rad2deg(heading),
            rotation_point="center",
            alpha=0.8,
            zorder=30,
            facecolor=color,
            label=f"Vehicle {self._vehicle_id}",
        )

        ax.add_patch(rect_patch)

        # Plot trajectory
        if include_trace:
            if timestep is None:
                ax.plot(
                    self._x, self._y, "-", color=rect_patch.get_facecolor(), zorder=31
                )
            else:
                ax.plot(
                    self._x[timestep:],
                    self._y[timestep:],
                    "-",
                    color=rect_patch.get_facecolor(),
                    zorder=31,
                )
        elif timestep is None:
            ax.plot(self._x, self._y, "x", color=rect_patch.get_facecolor(), zorder=31)

    def _format_ax(self, ax: plt.Axes, *args, **kwargs) -> None:  # noqa: ARG002
        margin = 5
        ax.set_xlim(np.min(self._x) - margin, np.max(self._x) + margin)
        ax.set_ylim(np.min(self._y) - margin, np.max(self._y) + margin)

        ax.set_aspect("equal")

        ax.set(title="Vehicle", xlabel="X position in m", ylabel="Y position in m")



import numpy as np
import warnings

from commonroad.scenario.lanelet import (
    LaneletNetwork,
    Lanelet,
    LaneletType,
    LineMarking,
)
from commonroad.scenario.traffic_sign import TrafficSignElement, TrafficSignIDZamunda
from commonroad_dc.pycrccosy import CurvilinearCoordinateSystem
from shapely.errors import ShapelyDeprecationWarning

from .traffic_sign_fixed import TraffiSignFixed
from ..road.road import Road


class LaneletNetworkCreator:
    def __init__(self, road: Road) -> None:
        self._lanelet_network = self._create_lanelet_network(road)

    @property
    def lanelet_network(self) -> LaneletNetwork:
        return self._lanelet_network

    def _create_lanelet_network(self, road: Road) -> LaneletNetwork:
        """
        Generate CR-LaneletNetwork from Road object
        """

        n_lanes = road.n_lanes
        all_lines = road.offset_lines
        all_lines[0.0] = road.ref_line

        offsets = sorted(all_lines.keys(), reverse=True)

        lanelet_list = []

        # CAUTION: Do not use lanelet ID 0. This will break the adj_left, adj_right relations when writing to xml file
        # see class LaneletXMLNode code in CRFileWriter
        lanelet_min_id = 1000
        lanelet_ids = range(lanelet_min_id, n_lanes + lanelet_min_id)

        for i_lanelet, lanelet_id in enumerate(lanelet_ids):
            # Neighbours
            if (lanelet_id - 1) in lanelet_ids:
                adj_right = lanelet_id - 1
                adj_right_direction = True
                right_marking = LineMarking.DASHED
            else:
                adj_right = None
                adj_right_direction = False
                right_marking = LineMarking.SOLID

            if (lanelet_id + 1) in lanelet_ids:
                adj_left = lanelet_id + 1
                adj_left_direction = True
                left_marking = LineMarking.DASHED
            else:
                adj_left = None
                adj_left_direction = False
                left_marking = LineMarking.SOLID

            # (Virtual) speed limit sign
            speed_limit_traffic_sign = TraffiSignFixed(
                max(lanelet_ids) + 100,
                [
                    TrafficSignElement(
                        TrafficSignIDZamunda["MAX_SPEED"], [str(road.speed_limit)]
                    )
                ],
                {lanelet_id},
                position=np.array([0, 0]),
                virtual=True,
            )

            # Select lines
            left_line_offset = offsets[2 + 2 * i_lanelet]
            center_line_offset = offsets[1 + 2 * i_lanelet]
            right_line_offset = offsets[2 * i_lanelet]

            left_vertices = all_lines[left_line_offset]
            center_vertices = all_lines[center_line_offset]
            right_vertices = all_lines[right_line_offset]

            # Create lanelet object
            lanelet = Lanelet(
                left_vertices,
                center_vertices,
                right_vertices,
                lanelet_id,
                adjacent_left=adj_left,
                adjacent_left_same_direction=adj_left_direction,
                adjacent_right=adj_right,
                adjacent_right_same_direction=adj_right_direction,
                line_marking_left_vertices=left_marking,
                line_marking_right_vertices=right_marking,
                lanelet_type={LaneletType.HIGHWAY},
                traffic_signs={speed_limit_traffic_sign},
            )

            lanelet_list.append(lanelet)

        # Create lanelet
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", category=ShapelyDeprecationWarning)
            lanelet_network = LaneletNetwork.create_from_lanelet_list(lanelet_list)

        # Add traffic signs
        for lanelet in lanelet_list:
            for ts in lanelet.traffic_signs:
                # Catch warning that the traffic light is added multiple times to each lanelet.
                with warnings.catch_warnings():
                    warnings.simplefilter("ignore", category=UserWarning)
                    lanelet_network.add_traffic_sign(
                        ts, set()
                    )  # references already existant in lanelet, but traffic sign object missing

        return lanelet_network

    @staticmethod
    def arc_heading_at_s(
        curvilinear_cs: CurvilinearCoordinateSystem, s: float
    ) -> float:
        """Returns tangent angle at s in rad"""

        tanget_vector = curvilinear_cs.tangent(s)
        orientation = np.arctan2(tanget_vector[1], tanget_vector[0])

        return orientation



from pathlib import Path

from simple_scenario import Scenario, EgoConfiguration, Vehicle
from simple_scenario.road import Road, StraightSegment


class TestOncomingTraffic:
    RESULT_DIR = Path(__file__).parent / "results" / "test_oncoming_traffic"
    RESULT_DIR.mkdir(exist_ok=True)

    def test_scenario_gen(self):
        result_dir = self.RESULT_DIR / "test_scenario_gen"
        result_dir.mkdir(exist_ok=True)

        road = Road(2, 3.75, [StraightSegment(500)])

        ego = EgoConfiguration(1000, 50, 0, 100 / 3.6)

        overtaking_vehicle = Vehicle(
            0,
            1001,
            400,
            0,
            100 / 3.6,
            lc_delay=3,
            lc_direction=-1,
            inverse_driving_direction=True,
        )
        slow_vehicle = Vehicle(
            1, 1001, 330, 0.5, 60 / 3.6, inverse_driving_direction=True
        )

        scenario = Scenario(
            "oncoming_traffic", road, ego, [overtaking_vehicle, slow_vehicle], 13
        )

        scenario.render_gif(result_dir)

        for vehicle in scenario.vehicles:
            if vehicle.inverse_driving_direction:
                assert vehicle.x[-1] < vehicle.x[0], (
                    f"Vehicle {vehicle.id} should drive in inverse driving direction, but it is not."
                )

        print("TEST PASSED")


if __name__ == "__main__":
    tester = TestOncomingTraffic()
    tester.test_scenario_gen()



import matplotlib.pyplot as plt
import pytest

from pathlib import Path

from simple_scenario import LXD_AVAILABLE

if LXD_AVAILABLE:
    from lxd_io import Dataset

from simple_scenario.road import Road


class TestHighdRoadCreation:
    DATA_DIR = Path(__file__).parent / "assets/highD-dataset-v1.0"

    RESULT_DIR = Path(__file__).parent / "results" / "test_highd_road_creation"
    RESULT_DIR.mkdir(exist_ok=True)

    @pytest.mark.skipif(not LXD_AVAILABLE, reason="lxd extra is not installed")
    def test_highd_road_creation(self):
        result_dir = self.RESULT_DIR
        dataset_dir = self.DATA_DIR

        dataset = Dataset(dataset_dir)

        for recording_id in dataset.recording_ids:
            print(f"recording: {recording_id:02d} / {len(dataset.recording_ids):02d}")

            recording = dataset.get_recording(recording_id)

            speed_limit = recording.get_meta_data("speedLimit")
            if speed_limit == -1:
                speed_limit = 120

            lower_road = Road.from_highd_parameters(
                recording.get_meta_data("lowerLaneMarkings"),
                "lower",
                speed_limit=speed_limit,
            )
            upper_road = Road.from_highd_parameters(
                recording.get_meta_data("upperLaneMarkings"),
                "upper",
                speed_limit=speed_limit,
            )

            recording_background = recording.get_background_image()

            px2m_factor = 0.1
            background_image_scale_factor = dataset._background_image_scale_factor  # noqa: SLF001

            roads = {"lower": lower_road, "upper": upper_road}

            for road_name, road in roads.items():
                road.render(
                    ".",
                    str(result_dir / f"{recording_id:02d}_{road_name}_lanelets"),
                    verbose=True,
                )

                ref_line_in_image = (
                    road.ref_line / px2m_factor / background_image_scale_factor
                )
                ref_line_in_image[:, 1] = -ref_line_in_image[:, 1]

                height, width, _ = recording_background.shape
                dpi = 600
                f, ax = plt.subplots(figsize=(width / dpi, height / dpi), dpi=dpi)
                ax.imshow(recording_background)

                ax.plot(ref_line_in_image[:, 0], ref_line_in_image[:, 1], "r-", lw=0.1)

                for offset_line in road.offset_lines.values():
                    offset_line_in_image = (
                        offset_line / px2m_factor / background_image_scale_factor
                    )
                    offset_line_in_image[:, 1] = -offset_line_in_image[:, 1]

                    ax.plot(
                        offset_line_in_image[:, 0],
                        offset_line_in_image[:, 1],
                        "b-",
                        lw=0.1,
                    )

                ax.axis("off")
                f.savefig(
                    result_dir / f"{recording_id:02d}_{road_name}_on_background.jpg",
                    bbox_inches="tight",
                    pad_inches=0,
                )

                plt.close(f)


if __name__ == "__main__":
    tester = TestHighdRoadCreation()
    tester.test_highd_road_creation()



"""
This file is a modified version of the following file: https://gitlab.lrz.de/tum-cps/commonroad_io/-/blob/cc4c594c9d5409a5621535e804b58bdacdf533f6/commonroad/scenario/traffic_sign_interpreter.py

The original file (see above) is licensed under the following license:

BSD 3-Clause License

Copyright 2021 Technical University of Munich, Professorship of Cyber-Physical Systems.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice,
   this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice,
   this list of conditions and the following disclaimer in the documentation
   and/or other materials provided with the distribution.

3. Neither the name of the copyright holder nor the names of its
   contributors may be used to endorse or promote products derived from
   this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""

from functools import lru_cache
from typing import Union, FrozenSet
from commonroad.scenario.traffic_sign_interpreter import TrafficSigInterpreter


class TrafficSigInterpreterFixed(TrafficSigInterpreter):
    """
    Fix speed limit look up (expects ID instead of traffic sign, but this is only available in the scenario if it is saved once to XML before).
    During creation, the TrafficSign is needed..
    """

    @lru_cache(maxsize=1024)
    def speed_limit(self, lanelet_ids: FrozenSet[int]) -> Union[float, None]:
        """
        Extracts the maximum speed limit of provided lanelets

        :param lanelet_ids: set of lanelets which should be considered
        :returns: speed limit of provided lanelets or None if no speed limit exists

        """
        speed_limits = []
        for lanelet_id in lanelet_ids:
            lanelet = self._lanelet_network.find_lanelet_by_id(lanelet_id)
            for traffic_sign_id in lanelet.traffic_signs:

                # --- THE FIX ---
                traffic_sign_id_ = traffic_sign_id
                if not isinstance(traffic_sign_id, int):
                    # It is a TrafficSign object
                    traffic_sign_id_ = traffic_sign_id.traffic_sign_id
                # --- THE FIX ---

                traffic_sign = self._lanelet_network.find_traffic_sign_by_id(traffic_sign_id_)
                for elem in traffic_sign.traffic_sign_elements:
                    if elem.traffic_sign_element_id == self.traffic_sign_ids.MAX_SPEED:
                        speed_limits.append(float(elem.additional_values[0]))

        if len(speed_limits) == 0:
            speed_limit = None
        else:
            speed_limit = min(speed_limits)
        return speed_limit
